local addonName, addonTable = ...

local tracker = LibStub("LibStormEarthAndFire-1.0")

local printHeader = "|cff00aa00Storm, Earth, and Fire|r"

local function debug(...)
	local arg = ...
	print("|cff00aa00Storm, Earth, and Fire |cffff0000debug|r: " .. tostring(arg), select(2, ...))
end

local addon = {}

--[====[ Config ]====]

addon.DB = addonTable.DB

addon.DB:SetDefaults({
	configMode = true,
	growDown = true,
	playSoundOnMaster = true,
})

addon.DB:RegisterCallback(addon, "configMode", function(key, oldValue, newValue)
	if addon.enabled then
		addon:updateConfigFrames()
		addon:relinkFrames()
	end
end)

addon.DB:RegisterCallback(addon, "growDown", function(key, oldValue, newValue)
	if addon.enabled then
		addon:relinkFrames()
	end
end)

local handleConfigMsg
do
	local boolConfig = { "configMode", "growDown", "playSoundOnMaster" }
	function handleConfigMsg(tokens)
		if tokens[1] == "list" then
			if #tokens ~= 1 then
				return false
			end
			print(printHeader.." config:")
			for _, key in ipairs(boolConfig) do
				print("    " .. key .. " - " .. tostring(addon.DB[key]))
			end
			return true
		end
		if tokens[1] == "set" then
			if #tokens ~= 3 then
				return false
			end
			local key, value = tokens[2]:lower(), tokens[3]
			local ok = false
			for _, v in ipairs(boolConfig) do
				if key == v:lower() then
					key = v
					ok = true
					break
				end
			end
			if ok then
				if value:lower() == "true" then
					addon.DB[key] = true
				elseif value:lower() == "false" then
					addon.DB[key] = false
				else
					error("config key " .. key .. " is a boolean")
				end
			else
				error("unknown key " .. key)
			end
		else
			if #tokens ~= 1 then
				return false
			end
			local key = tokens[1]:lower()
			local ok = false
			for _, v in ipairs(boolConfig) do
				if key == v:lower() then
					key = v
					ok = true
					break
				end
			end
			if ok then
				print(printHeader.." config: " .. key .. " - " .. tostring(addon.DB[key]))
			else
				error("unknown key " .. key)
			end
		end
		return true
	end
end

--[====[ Implementation ]====]

local frames = { StormEarthAndFireFrame1, StormEarthAndFireFrame2 }

local UnitTracker = addonTable.UnitTracker

addon.enabled = false
function addon:Enable()
	if self.enabled then return end
	self.enabled = true
	self:updateConfigFrames()
	for _, frame in ipairs(frames) do
		if frame.guid then
			frame:Show()
		end
	end
	self:UpdateRaidIcons()
	self:relinkFrames()
end

function addon:Disable()
	if not self.enabled then return end
	self.enabled = false
	self:updateConfigFrames()
	for _, frame in ipairs(frames) do
		frame:Hide()
	end
end

local function getNPCID(guid)
	if tonumber(guid:sub(5,5), 16) % 8 == 3 then
		return tonumber(guid:sub(6,10), 16)
	end
	return nil
end

local active = {}
function addon:displaySpirit(name, guid, target, targetGUID, icon)
	local frame
	for _, f in ipairs(frames) do
		if not f.guid then
			frame = f
			break
		end
	end
	if not frame then
		error("Couldn't find available frame for spirit")
	end
	frame.label:SetText(target)
	frame.label:SetVertexColor(1.0, 1.0, 1.0)
	frame:SizeToFit()
	frame.icon:SetTexture(icon)
	SetDesaturation(frame.icon, false)

	frame.raidIcon:Hide()

	frame.guid = guid
	frame.targetGUID = targetGUID

	table.insert(active, frame)
	self:UpdateRaidIcons()
	self:relinkFrames()

	frame:Show()
end

function addon:hideSpirit(name, guid)
	local frame, idx
	for i, f in ipairs(active) do
		if f.guid == guid then
			frame = f
			idx = i
			break
		end
	end
	if not frame then return end
	frame:Hide()
	frame.guid = nil
	frame:ResetBackdropColor()
	frame:ClearAllPoints()
	table.remove(active, idx)
	PlaySoundFile("Sound\\Doodad\\Blackrockdoorsingledestroy.Ogg", self.DB.playSoundOnMaster and "master" or "SFX")
	if self.DB.configMode then
		self:updateConfigFrames()
	end
	self:relinkFrames()
end

function addon:relinkFrames()
	local last = nil
	local point, relPoint
	if self.DB.growDown then
		point, relPoint = "TOPLEFT", "BOTTOMLEFT"
	else
		point, relPoint = "BOTTOMLEFT", "TOPLEFT"
	end
	local function relink(frame)
		frame:ClearAllPoints()
		if last then
			frame:SetPoint(point, last,relPoint)
		else
			frame:SetPoint(point, StormEarthAndFireAnchor, "CENTER")
		end
		last = frame
	end
	for _, frame in ipairs(active) do
		relink(frame)
	end
	if self.DB.configMode then
		for _, frame in ipairs(frames) do
			if not frame.guid then
				relink(frame)
			end
		end
	end
end

function addon:warnSpiritTarget(name, guid, target, targetGUID)
	for _, frame in ipairs(frames) do
		if frame.guid == guid then
			frame:SetBackdropColor(0.8, 0.09, 0.09)
		else
			frame:ResetBackdropColor()
		end
	end
end

function addon:cancelWarnSpiritTarget()
	for _, frame in ipairs(frames) do
		frame:ResetBackdropColor()
	end
end

do
	local configTextures = {
		"Interface\\Icons\\INV_SummerFest_FireSpirit",
		"Interface\\Icons\\inv_misc_stormlordsfavor",
		"Interface\\Icons\\INV_Elemental_Primal_Earth"
	}

	function addon:updateConfigFrames()
		if self.DB.configMode and self.enabled then
			StormEarthAndFireAnchor:Show()
			local used = {}
			for _, frame in ipairs(frames) do
				if frame.guid then
					used[frame.icon:GetTexture():lower()] = true
				end
			end
			for _, frame in ipairs(frames) do
				if not frame.guid then
					frame.label:SetText((GetSpellInfo(137639)))
					frame.label:SetVertexColor(0.5, 0.5, 0.5)
					frame:SizeToFit()
					for _, tex in ipairs(configTextures) do
						if not used[tex:lower()] then
							frame.icon:SetTexture(tex)
							used[tex:lower()] = true
							break
						end
					end
					SetDesaturation(frame.icon, true)
					frame.raidIcon():Hide()
					frame:Show()
				end
			end
		else
			StormEarthAndFireAnchor:Hide()
			for _, frame in ipairs(frames) do
				if not frame.guid then
					frame:ClearAllPoints()
					frame:Hide()
				end
			end
		end
	end
end

function addon:UpdateRaidIcons()
	for _, frame in ipairs(frames) do
		if frame.guid then
			local unit = UnitTracker:UnitIDForGUID(frame.targetGUID)
			local index
			if unit then index = GetRaidTargetIndex(unit) end
			if index then
				SetRaidTargetIconTexture(frame.raidIcon, index)
				frame.raidIcon:SetDim(false)
				frame.raidIcon:Show()
			elseif unit and (unit ~= "mouseover" or UnitGUID(unit) == frame.targetGUID) then
				frame.raidIcon:Hide()
			else
				frame.raidIcon:SetDim(true)
			end
		end
	end
end

--[====[ Slash Commands ]====]

do
	local function tokenize(str)
		local tokens = {}
		for word in str:gmatch("%S+") do
			table.insert(tokens, word)
		end
		return tokens
	end
	SlashCmdList[addonName:upper()] = function(msg, editBox)
		local tokens = tokenize(msg)
		if tokens[1] == "config" then
			table.remove(tokens, 1)
			local ok, success = pcall(handleConfigMsg, tokens)
			if ok and not success then
				print(printHeader.." config usage:")
				print("    config list - view all config keys and values")
				print("    config <key> - view value for <key>")
				print("    config set <key> <value> - set <key> to <value>")
			elseif not ok then
				print(printHeader.." |cffff0000error|r: " .. success:gsub("^.-: ", "", 1))
			end
			return
		end
		print(printHeader.." commands:")
		print("  config - get/set config variables")
	end
	_G["SLASH_"..addonName:upper().."1"] = "/"..addonName
	_G["SLASH_"..addonName:upper().."2"] = "/sef"
end

--[====[ Events ]====]

function addon:TRACKER_ENABLED(event)
	self:Enable()
end
function addon:TRACKER_DISABLED(event)
	self:Disable()
end
do
	local spiritTargets = {}
	function addon:TRACKER_NEW_SPIRIT(event, name, guid, target, targetGUID, icon)
		local unit = UnitTracker:StartTrackingGUID(targetGUID)
		self:displaySpirit(name, guid, target, targetGUID, icon)
		spiritTargets[guid] = targetGUID
		if unit then
			self:UNIT_ID_AVAILABLE(nil, targetGUID, unit)
		end
	end
	function addon:TRACKER_DEAD_SPIRIT(event, name, guid)
		self:hideSpirit(name, guid)
		UnitTracker:StopTrackingGUID(spiritTargets[guid])
		spiritTargets[guid] = nil
	end
end
function addon:TRACKER_SPIRIT_TARGET(event, name, guid, target, targetGUID)
	self:warnSpiritTarget(name, guid, target, targetGUID)
end
function addon:TRACKER_SPIRIT_TARGET_END(event)
	self:cancelWarnSpiritTarget()
end

function addon:UNIT_ID_AVAILABLE(event, guid, unitid)
	self:UpdateRaidIcons()
end
function addon:UNIT_ID_GONE(event, guid)
	self:UpdateRaidIcons()
end

for _, v in ipairs({"ENABLED", "DISABLED", "NEW_SPIRIT", "DEAD_SPIRIT", "SPIRIT_TARGET", "SPIRIT_TARGET_END"}) do
	tracker.RegisterCallback(addon, v, "TRACKER_"..v)
end

for _, v in ipairs({"UNIT_ID_AVAILABLE", "UNIT_ID_GONE"}) do
	UnitTracker.RegisterCallback(addon, v)
end

local frame = CreateFrame("frame")
frame:RegisterEvent("ADDON_LOADED")
frame:RegisterEvent("PLAYER_ENTERING_WORLD")
frame:SetScript("OnEvent", function(self, event, ...)
	addon["EVENT_"..event](addon, ...)
end)

function addon:EVENT_ADDON_LOADED(name)
	if name == addonName then
		self.DB:Initialize()
		frame:UnregisterEvent("ADDON_LOADED")
	end
end
function addon:EVENT_PLAYER_ENTERING_WORLD()
	-- It seems GameFontNormalSmall may change between ADDON_LOADED and now
	self:updateConfigFrames()
	frame:UnregisterEvent("PLAYER_ENTERING_WORLD")
end

function addon:EVENT_RAID_TARGET_UPDATE()
	self:UpdateRaidIcons()
end
frame:RegisterEvent("RAID_TARGET_UPDATE")
